# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from metronome import Metronome, AsyncMetronome
from tests.utils import assert_matches_type
from metronome._utils import parse_datetime
from metronome.types.v1 import (
    CustomerDetail,
    CustomerCreateResponse,
    CustomerArchiveResponse,
    CustomerSetNameResponse,
    CustomerRetrieveResponse,
    CustomerListCostsResponse,
    CustomerPreviewEventsResponse,
    CustomerListBillableMetricsResponse,
)
from metronome.pagination import SyncCursorPage, AsyncCursorPage

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestCustomers:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_create(self, client: Metronome) -> None:
        customer = client.v1.customers.create(
            name="Example, Inc.",
        )
        assert_matches_type(CustomerCreateResponse, customer, path=["response"])

    @parametrize
    def test_method_create_with_all_params(self, client: Metronome) -> None:
        customer = client.v1.customers.create(
            name="Example, Inc.",
            billing_config={
                "billing_provider_customer_id": "billing_provider_customer_id",
                "billing_provider_type": "aws_marketplace",
                "aws_is_subscription_product": True,
                "aws_product_code": "aws_product_code",
                "aws_region": "af-south-1",
                "stripe_collection_method": "charge_automatically",
            },
            custom_fields={"foo": "string"},
            customer_billing_provider_configurations=[
                {
                    "billing_provider": "stripe",
                    "configuration": {
                        "stripe_customer_id": "bar",
                        "stripe_collection_method": "bar",
                    },
                    "delivery_method": "direct_to_billing_provider",
                    "delivery_method_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                }
            ],
            external_id="x",
            ingest_aliases=["team@example.com"],
        )
        assert_matches_type(CustomerCreateResponse, customer, path=["response"])

    @parametrize
    def test_raw_response_create(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.create(
            name="Example, Inc.",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(CustomerCreateResponse, customer, path=["response"])

    @parametrize
    def test_streaming_response_create(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.create(
            name="Example, Inc.",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(CustomerCreateResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_retrieve(self, client: Metronome) -> None:
        customer = client.v1.customers.retrieve(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )
        assert_matches_type(CustomerRetrieveResponse, customer, path=["response"])

    @parametrize
    def test_raw_response_retrieve(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.retrieve(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(CustomerRetrieveResponse, customer, path=["response"])

    @parametrize
    def test_streaming_response_retrieve(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.retrieve(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(CustomerRetrieveResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve(self, client: Metronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.v1.customers.with_raw_response.retrieve(
                customer_id="",
            )

    @parametrize
    def test_method_list(self, client: Metronome) -> None:
        customer = client.v1.customers.list()
        assert_matches_type(SyncCursorPage[CustomerDetail], customer, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Metronome) -> None:
        customer = client.v1.customers.list(
            customer_ids=["string"],
            ingest_alias="ingest_alias",
            limit=1,
            next_page="next_page",
            only_archived=True,
            salesforce_account_ids=["string"],
        )
        assert_matches_type(SyncCursorPage[CustomerDetail], customer, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(SyncCursorPage[CustomerDetail], customer, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(SyncCursorPage[CustomerDetail], customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_archive(self, client: Metronome) -> None:
        customer = client.v1.customers.archive(
            id="8deed800-1b7a-495d-a207-6c52bac54dc9",
        )
        assert_matches_type(CustomerArchiveResponse, customer, path=["response"])

    @parametrize
    def test_raw_response_archive(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.archive(
            id="8deed800-1b7a-495d-a207-6c52bac54dc9",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(CustomerArchiveResponse, customer, path=["response"])

    @parametrize
    def test_streaming_response_archive(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.archive(
            id="8deed800-1b7a-495d-a207-6c52bac54dc9",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(CustomerArchiveResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_list_billable_metrics(self, client: Metronome) -> None:
        customer = client.v1.customers.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )
        assert_matches_type(SyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

    @parametrize
    def test_method_list_billable_metrics_with_all_params(self, client: Metronome) -> None:
        customer = client.v1.customers.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            include_archived=True,
            limit=1,
            next_page="next_page",
            on_current_plan=True,
        )
        assert_matches_type(SyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

    @parametrize
    def test_raw_response_list_billable_metrics(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(SyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

    @parametrize
    def test_streaming_response_list_billable_metrics(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(SyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_list_billable_metrics(self, client: Metronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.v1.customers.with_raw_response.list_billable_metrics(
                customer_id="",
            )

    @parametrize
    def test_method_list_costs(self, client: Metronome) -> None:
        customer = client.v1.customers.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(SyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

    @parametrize
    def test_method_list_costs_with_all_params(self, client: Metronome) -> None:
        customer = client.v1.customers.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
            limit=1,
            next_page="next_page",
        )
        assert_matches_type(SyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

    @parametrize
    def test_raw_response_list_costs(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(SyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

    @parametrize
    def test_streaming_response_list_costs(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(SyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_list_costs(self, client: Metronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.v1.customers.with_raw_response.list_costs(
                customer_id="",
                ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
                starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
            )

    @parametrize
    def test_method_preview_events(self, client: Metronome) -> None:
        customer = client.v1.customers.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[{"event_type": "heartbeat"}],
        )
        assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

    @parametrize
    def test_method_preview_events_with_all_params(self, client: Metronome) -> None:
        customer = client.v1.customers.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[
                {
                    "event_type": "heartbeat",
                    "customer_id": "x",
                    "properties": {
                        "cpu_hours": "bar",
                        "memory_gb_hours": "bar",
                    },
                    "timestamp": "2021-01-01T00:00:00Z",
                    "transaction_id": "x",
                }
            ],
            mode="replace",
            skip_zero_qty_line_items=True,
        )
        assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

    @parametrize
    def test_raw_response_preview_events(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[{"event_type": "heartbeat"}],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

    @parametrize
    def test_streaming_response_preview_events(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[{"event_type": "heartbeat"}],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_preview_events(self, client: Metronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.v1.customers.with_raw_response.preview_events(
                customer_id="",
                events=[{"event_type": "heartbeat"}],
            )

    @parametrize
    def test_method_set_ingest_aliases(self, client: Metronome) -> None:
        customer = client.v1.customers.set_ingest_aliases(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ingest_aliases=["team@example.com"],
        )
        assert customer is None

    @parametrize
    def test_raw_response_set_ingest_aliases(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.set_ingest_aliases(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ingest_aliases=["team@example.com"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert customer is None

    @parametrize
    def test_streaming_response_set_ingest_aliases(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.set_ingest_aliases(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ingest_aliases=["team@example.com"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert customer is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_set_ingest_aliases(self, client: Metronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.v1.customers.with_raw_response.set_ingest_aliases(
                customer_id="",
                ingest_aliases=["team@example.com"],
            )

    @parametrize
    def test_method_set_name(self, client: Metronome) -> None:
        customer = client.v1.customers.set_name(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            name="Example, Inc.",
        )
        assert_matches_type(CustomerSetNameResponse, customer, path=["response"])

    @parametrize
    def test_raw_response_set_name(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.set_name(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            name="Example, Inc.",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert_matches_type(CustomerSetNameResponse, customer, path=["response"])

    @parametrize
    def test_streaming_response_set_name(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.set_name(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            name="Example, Inc.",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert_matches_type(CustomerSetNameResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_set_name(self, client: Metronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.v1.customers.with_raw_response.set_name(
                customer_id="",
                name="Example, Inc.",
            )

    @parametrize
    def test_method_update_config(self, client: Metronome) -> None:
        customer = client.v1.customers.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )
        assert customer is None

    @parametrize
    def test_method_update_config_with_all_params(self, client: Metronome) -> None:
        customer = client.v1.customers.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            leave_stripe_invoices_in_draft=True,
            salesforce_account_id="0015500001WO1ZiABL",
        )
        assert customer is None

    @parametrize
    def test_raw_response_update_config(self, client: Metronome) -> None:
        response = client.v1.customers.with_raw_response.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = response.parse()
        assert customer is None

    @parametrize
    def test_streaming_response_update_config(self, client: Metronome) -> None:
        with client.v1.customers.with_streaming_response.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = response.parse()
            assert customer is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update_config(self, client: Metronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            client.v1.customers.with_raw_response.update_config(
                customer_id="",
            )


class TestAsyncCustomers:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_create(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.create(
            name="Example, Inc.",
        )
        assert_matches_type(CustomerCreateResponse, customer, path=["response"])

    @parametrize
    async def test_method_create_with_all_params(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.create(
            name="Example, Inc.",
            billing_config={
                "billing_provider_customer_id": "billing_provider_customer_id",
                "billing_provider_type": "aws_marketplace",
                "aws_is_subscription_product": True,
                "aws_product_code": "aws_product_code",
                "aws_region": "af-south-1",
                "stripe_collection_method": "charge_automatically",
            },
            custom_fields={"foo": "string"},
            customer_billing_provider_configurations=[
                {
                    "billing_provider": "stripe",
                    "configuration": {
                        "stripe_customer_id": "bar",
                        "stripe_collection_method": "bar",
                    },
                    "delivery_method": "direct_to_billing_provider",
                    "delivery_method_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                }
            ],
            external_id="x",
            ingest_aliases=["team@example.com"],
        )
        assert_matches_type(CustomerCreateResponse, customer, path=["response"])

    @parametrize
    async def test_raw_response_create(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.create(
            name="Example, Inc.",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(CustomerCreateResponse, customer, path=["response"])

    @parametrize
    async def test_streaming_response_create(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.create(
            name="Example, Inc.",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(CustomerCreateResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_retrieve(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.retrieve(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )
        assert_matches_type(CustomerRetrieveResponse, customer, path=["response"])

    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.retrieve(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(CustomerRetrieveResponse, customer, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.retrieve(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(CustomerRetrieveResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncMetronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.v1.customers.with_raw_response.retrieve(
                customer_id="",
            )

    @parametrize
    async def test_method_list(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.list()
        assert_matches_type(AsyncCursorPage[CustomerDetail], customer, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.list(
            customer_ids=["string"],
            ingest_alias="ingest_alias",
            limit=1,
            next_page="next_page",
            only_archived=True,
            salesforce_account_ids=["string"],
        )
        assert_matches_type(AsyncCursorPage[CustomerDetail], customer, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(AsyncCursorPage[CustomerDetail], customer, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(AsyncCursorPage[CustomerDetail], customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_archive(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.archive(
            id="8deed800-1b7a-495d-a207-6c52bac54dc9",
        )
        assert_matches_type(CustomerArchiveResponse, customer, path=["response"])

    @parametrize
    async def test_raw_response_archive(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.archive(
            id="8deed800-1b7a-495d-a207-6c52bac54dc9",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(CustomerArchiveResponse, customer, path=["response"])

    @parametrize
    async def test_streaming_response_archive(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.archive(
            id="8deed800-1b7a-495d-a207-6c52bac54dc9",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(CustomerArchiveResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_list_billable_metrics(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )
        assert_matches_type(AsyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

    @parametrize
    async def test_method_list_billable_metrics_with_all_params(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            include_archived=True,
            limit=1,
            next_page="next_page",
            on_current_plan=True,
        )
        assert_matches_type(AsyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

    @parametrize
    async def test_raw_response_list_billable_metrics(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(AsyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

    @parametrize
    async def test_streaming_response_list_billable_metrics(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.list_billable_metrics(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(AsyncCursorPage[CustomerListBillableMetricsResponse], customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_list_billable_metrics(self, async_client: AsyncMetronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.v1.customers.with_raw_response.list_billable_metrics(
                customer_id="",
            )

    @parametrize
    async def test_method_list_costs(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(AsyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

    @parametrize
    async def test_method_list_costs_with_all_params(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
            limit=1,
            next_page="next_page",
        )
        assert_matches_type(AsyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

    @parametrize
    async def test_raw_response_list_costs(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(AsyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

    @parametrize
    async def test_streaming_response_list_costs(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.list_costs(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(AsyncCursorPage[CustomerListCostsResponse], customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_list_costs(self, async_client: AsyncMetronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.v1.customers.with_raw_response.list_costs(
                customer_id="",
                ending_before=parse_datetime("2019-12-27T18:11:19.117Z"),
                starting_on=parse_datetime("2019-12-27T18:11:19.117Z"),
            )

    @parametrize
    async def test_method_preview_events(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[{"event_type": "heartbeat"}],
        )
        assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

    @parametrize
    async def test_method_preview_events_with_all_params(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[
                {
                    "event_type": "heartbeat",
                    "customer_id": "x",
                    "properties": {
                        "cpu_hours": "bar",
                        "memory_gb_hours": "bar",
                    },
                    "timestamp": "2021-01-01T00:00:00Z",
                    "transaction_id": "x",
                }
            ],
            mode="replace",
            skip_zero_qty_line_items=True,
        )
        assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

    @parametrize
    async def test_raw_response_preview_events(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[{"event_type": "heartbeat"}],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

    @parametrize
    async def test_streaming_response_preview_events(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.preview_events(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            events=[{"event_type": "heartbeat"}],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(CustomerPreviewEventsResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_preview_events(self, async_client: AsyncMetronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.v1.customers.with_raw_response.preview_events(
                customer_id="",
                events=[{"event_type": "heartbeat"}],
            )

    @parametrize
    async def test_method_set_ingest_aliases(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.set_ingest_aliases(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ingest_aliases=["team@example.com"],
        )
        assert customer is None

    @parametrize
    async def test_raw_response_set_ingest_aliases(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.set_ingest_aliases(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ingest_aliases=["team@example.com"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert customer is None

    @parametrize
    async def test_streaming_response_set_ingest_aliases(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.set_ingest_aliases(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            ingest_aliases=["team@example.com"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert customer is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_set_ingest_aliases(self, async_client: AsyncMetronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.v1.customers.with_raw_response.set_ingest_aliases(
                customer_id="",
                ingest_aliases=["team@example.com"],
            )

    @parametrize
    async def test_method_set_name(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.set_name(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            name="Example, Inc.",
        )
        assert_matches_type(CustomerSetNameResponse, customer, path=["response"])

    @parametrize
    async def test_raw_response_set_name(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.set_name(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            name="Example, Inc.",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert_matches_type(CustomerSetNameResponse, customer, path=["response"])

    @parametrize
    async def test_streaming_response_set_name(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.set_name(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            name="Example, Inc.",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert_matches_type(CustomerSetNameResponse, customer, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_set_name(self, async_client: AsyncMetronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.v1.customers.with_raw_response.set_name(
                customer_id="",
                name="Example, Inc.",
            )

    @parametrize
    async def test_method_update_config(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )
        assert customer is None

    @parametrize
    async def test_method_update_config_with_all_params(self, async_client: AsyncMetronome) -> None:
        customer = await async_client.v1.customers.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
            leave_stripe_invoices_in_draft=True,
            salesforce_account_id="0015500001WO1ZiABL",
        )
        assert customer is None

    @parametrize
    async def test_raw_response_update_config(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.with_raw_response.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        customer = await response.parse()
        assert customer is None

    @parametrize
    async def test_streaming_response_update_config(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.with_streaming_response.update_config(
            customer_id="d7abd0cd-4ae9-4db7-8676-e986a4ebd8dc",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            customer = await response.parse()
            assert customer is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update_config(self, async_client: AsyncMetronome) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `customer_id` but received ''"):
            await async_client.v1.customers.with_raw_response.update_config(
                customer_id="",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["PlanListCustomersParams"]


class PlanListCustomersParams(TypedDict, total=False):
    plan_id: Required[str]

    limit: int
    """Max number of results that should be returned"""

    next_page: str
    """Cursor that indicates where the next page of results should start."""

    status: Literal["all", "active", "ended", "upcoming"]
    """Status of customers on a given plan. Defaults to `active`.

    - `all` - Return current, past, and upcoming customers of the plan.
    - `active` - Return current customers of the plan.
    - `ended` - Return past customers of the plan.
    - `upcoming` - Return upcoming customers of the plan.

    Multiple statuses can be OR'd together using commas, e.g. `active,ended`.
    **Note:** `ended,upcoming` combination is not yet supported.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional
from datetime import datetime

from ..._models import BaseModel

__all__ = ["CustomerListCostsResponse", "CreditTypes", "CreditTypesLineItemBreakdown"]


class CreditTypesLineItemBreakdown(BaseModel):
    cost: float

    name: str

    group_key: Optional[str] = None

    group_value: Optional[str] = None


class CreditTypes(BaseModel):
    cost: Optional[float] = None

    line_item_breakdown: Optional[List[CreditTypesLineItemBreakdown]] = None

    name: Optional[str] = None


class CustomerListCostsResponse(BaseModel):
    credit_types: Dict[str, CreditTypes]

    end_timestamp: datetime

    start_timestamp: datetime



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional
from datetime import datetime
from typing_extensions import Literal

from ...._models import BaseModel
from ...shared.rate import Rate
from ...shared.credit_type_data import CreditTypeData

__all__ = [
    "Invoice",
    "LineItem",
    "LineItemAppliedCommitOrCredit",
    "LineItemPostpaidCommit",
    "LineItemSubLineItem",
    "LineItemSubLineItemTierPeriod",
    "LineItemSubLineItemTier",
    "LineItemTier",
    "CorrectionRecord",
    "CorrectionRecordCorrectedExternalInvoice",
    "ExternalInvoice",
    "InvoiceAdjustment",
    "ResellerRoyalty",
    "ResellerRoyaltyAwsOptions",
    "ResellerRoyaltyGcpOptions",
]


class LineItemAppliedCommitOrCredit(BaseModel):
    id: str

    type: Literal["PREPAID", "POSTPAID", "CREDIT"]


class LineItemPostpaidCommit(BaseModel):
    id: str


class LineItemSubLineItemTierPeriod(BaseModel):
    starting_at: datetime

    ending_before: Optional[datetime] = None


class LineItemSubLineItemTier(BaseModel):
    price: float

    quantity: float

    starting_at: float
    """at what metric amount this tier begins"""

    subtotal: float


class LineItemSubLineItem(BaseModel):
    custom_fields: Dict[str, str]

    name: str

    quantity: float

    subtotal: float

    charge_id: Optional[str] = None

    credit_grant_id: Optional[str] = None

    end_date: Optional[datetime] = None
    """The end date for the charge (for seats charges only)."""

    price: Optional[float] = None
    """
    the unit price for this charge, present only if the charge is not tiered and the
    quantity is nonzero
    """

    start_date: Optional[datetime] = None
    """The start date for the charge (for seats charges only)."""

    tier_period: Optional[LineItemSubLineItemTierPeriod] = None
    """when the current tier started and ends (for tiered charges only)"""

    tiers: Optional[List[LineItemSubLineItemTier]] = None


class LineItemTier(BaseModel):
    level: float

    starting_at: str

    size: Optional[str] = None


class LineItem(BaseModel):
    credit_type: CreditTypeData

    name: str

    total: float

    applied_commit_or_credit: Optional[LineItemAppliedCommitOrCredit] = None
    """Details about the credit or commit that was applied to this line item.

    Only present on line items with product of `USAGE`, `SUBSCRIPTION` or
    `COMPOSITE` types.
    """

    commit_custom_fields: Optional[Dict[str, str]] = None

    commit_id: Optional[str] = None
    """
    For line items with product of `USAGE`, `SUBSCRIPTION`, or `COMPOSITE` types,
    the ID of the credit or commit that was applied to this line item. For line
    items with product type of `FIXED`, the ID of the prepaid or postpaid commit
    that is being paid for.
    """

    commit_netsuite_item_id: Optional[str] = None

    commit_netsuite_sales_order_id: Optional[str] = None

    commit_segment_id: Optional[str] = None

    commit_type: Optional[str] = None
    """
    `PrepaidCommit` (for commit types `PREPAID` and `CREDIT`) or `PostpaidCommit`
    (for commit type `POSTPAID`).
    """

    custom_fields: Optional[Dict[str, str]] = None

    discount_custom_fields: Optional[Dict[str, str]] = None

    discount_id: Optional[str] = None
    """ID of the discount applied to this line item."""

    ending_before: Optional[datetime] = None
    """The line item's end date (exclusive)."""

    group_key: Optional[str] = None

    group_value: Optional[str] = None

    is_prorated: Optional[bool] = None
    """Indicates whether the line item is prorated for `SUBSCRIPTION` type product."""

    list_price: Optional[Rate] = None
    """
    Only present for contract invoices and when the `include_list_prices` query
    parameter is set to true. This will include the list rate for the charge if
    applicable. Only present for usage and subscription line items.
    """

    metadata: Optional[str] = None

    netsuite_invoice_billing_end: Optional[datetime] = None
    """The end date for the billing period on the invoice."""

    netsuite_invoice_billing_start: Optional[datetime] = None
    """The start date for the billing period on the invoice."""

    netsuite_item_id: Optional[str] = None

    postpaid_commit: Optional[LineItemPostpaidCommit] = None
    """Only present for line items paying for a postpaid commit true-up."""

    presentation_group_values: Optional[Dict[str, Optional[str]]] = None
    """
    Includes the presentation group values associated with this line item if
    presentation group keys are used.
    """

    pricing_group_values: Optional[Dict[str, str]] = None
    """
    Includes the pricing group values associated with this line item if dimensional
    pricing is used.
    """

    product_custom_fields: Optional[Dict[str, str]] = None

    product_id: Optional[str] = None
    """ID of the product associated with the line item."""

    product_tags: Optional[List[str]] = None
    """The current product tags associated with the line item's `product_id`."""

    product_type: Optional[str] = None
    """The type of the line item's product.

    Possible values are `FixedProductListItem` (for `FIXED` type products),
    `UsageProductListItem` (for `USAGE` type products),
    `SubscriptionProductListItem` (for `SUBSCRIPTION` type products) or
    `CompositeProductListItem` (for `COMPOSITE` type products). For scheduled
    charges, commit and credit payments, the value is `FixedProductListItem`.
    """

    professional_service_custom_fields: Optional[Dict[str, str]] = None

    professional_service_id: Optional[str] = None

    quantity: Optional[float] = None
    """The quantity associated with the line item."""

    reseller_type: Optional[Literal["AWS", "AWS_PRO_SERVICE", "GCP", "GCP_PRO_SERVICE"]] = None

    scheduled_charge_custom_fields: Optional[Dict[str, str]] = None

    scheduled_charge_id: Optional[str] = None
    """ID of scheduled charge."""

    starting_at: Optional[datetime] = None
    """The line item's start date (inclusive)."""

    sub_line_items: Optional[List[LineItemSubLineItem]] = None

    subscription_custom_fields: Optional[Dict[str, str]] = None

    tier: Optional[LineItemTier] = None
    """Populated if the line item has a tiered price."""

    unit_price: Optional[float] = None
    """The unit price associated with the line item."""


class CorrectionRecordCorrectedExternalInvoice(BaseModel):
    billing_provider_type: Literal[
        "aws_marketplace",
        "stripe",
        "netsuite",
        "custom",
        "azure_marketplace",
        "quickbooks_online",
        "workday",
        "gcp_marketplace",
    ]

    external_status: Optional[
        Literal[
            "DRAFT",
            "FINALIZED",
            "PAID",
            "UNCOLLECTIBLE",
            "VOID",
            "DELETED",
            "PAYMENT_FAILED",
            "INVALID_REQUEST_ERROR",
            "SKIPPED",
            "SENT",
            "QUEUED",
        ]
    ] = None

    invoice_id: Optional[str] = None

    issued_at_timestamp: Optional[datetime] = None


class CorrectionRecord(BaseModel):
    corrected_invoice_id: str

    memo: str

    reason: str

    corrected_external_invoice: Optional[CorrectionRecordCorrectedExternalInvoice] = None


class ExternalInvoice(BaseModel):
    billing_provider_type: Literal[
        "aws_marketplace",
        "stripe",
        "netsuite",
        "custom",
        "azure_marketplace",
        "quickbooks_online",
        "workday",
        "gcp_marketplace",
    ]

    external_status: Optional[
        Literal[
            "DRAFT",
            "FINALIZED",
            "PAID",
            "UNCOLLECTIBLE",
            "VOID",
            "DELETED",
            "PAYMENT_FAILED",
            "INVALID_REQUEST_ERROR",
            "SKIPPED",
            "SENT",
            "QUEUED",
        ]
    ] = None

    invoice_id: Optional[str] = None

    issued_at_timestamp: Optional[datetime] = None


class InvoiceAdjustment(BaseModel):
    credit_type: CreditTypeData

    name: str

    total: float

    credit_grant_custom_fields: Optional[Dict[str, str]] = None

    credit_grant_id: Optional[str] = None


class ResellerRoyaltyAwsOptions(BaseModel):
    aws_account_number: Optional[str] = None

    aws_offer_id: Optional[str] = None

    aws_payer_reference_id: Optional[str] = None


class ResellerRoyaltyGcpOptions(BaseModel):
    gcp_account_id: Optional[str] = None

    gcp_offer_id: Optional[str] = None


class ResellerRoyalty(BaseModel):
    fraction: str

    netsuite_reseller_id: str

    reseller_type: Literal["AWS", "AWS_PRO_SERVICE", "GCP", "GCP_PRO_SERVICE"]

    aws_options: Optional[ResellerRoyaltyAwsOptions] = None

    gcp_options: Optional[ResellerRoyaltyGcpOptions] = None


class Invoice(BaseModel):
    id: str

    credit_type: CreditTypeData

    customer_id: str

    line_items: List[LineItem]

    status: str

    total: float

    type: str

    amendment_id: Optional[str] = None

    billable_status: Optional[Literal["billable", "unbillable"]] = None
    """This field's availability is dependent on your client's configuration."""

    contract_custom_fields: Optional[Dict[str, str]] = None

    contract_id: Optional[str] = None

    correction_record: Optional[CorrectionRecord] = None

    created_at: Optional[datetime] = None
    """When the invoice was created (UTC).

    This field is present for correction invoices only.
    """

    custom_fields: Optional[Dict[str, object]] = None

    customer_custom_fields: Optional[Dict[str, str]] = None

    end_timestamp: Optional[datetime] = None
    """End of the usage period this invoice covers (UTC)"""

    external_invoice: Optional[ExternalInvoice] = None

    invoice_adjustments: Optional[List[InvoiceAdjustment]] = None

    issued_at: Optional[datetime] = None
    """When the invoice was issued (UTC)"""

    net_payment_terms_days: Optional[float] = None

    netsuite_sales_order_id: Optional[str] = None
    """This field's availability is dependent on your client's configuration."""

    plan_custom_fields: Optional[Dict[str, str]] = None

    plan_id: Optional[str] = None

    plan_name: Optional[str] = None

    reseller_royalty: Optional[ResellerRoyalty] = None
    """Only present for contract invoices with reseller royalties."""

    salesforce_opportunity_id: Optional[str] = None
    """This field's availability is dependent on your client's configuration."""

    start_timestamp: Optional[datetime] = None
    """Beginning of the usage period this invoice covers (UTC)"""

    subtotal: Optional[float] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["CreditGrantVoidParams"]


class CreditGrantVoidParams(TypedDict, total=False):
    id: Required[str]

    release_uniqueness_key: bool
    """If true, resets the uniqueness key on this grant so it can be re-used"""

    void_credit_purchase_invoice: bool
    """If true, void the purchase invoice associated with the grant"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ...._models import BaseModel

__all__ = ["QuantityRounding"]


class QuantityRounding(BaseModel):
    decimal_places: float

    rounding_method: Literal["ROUND_UP", "ROUND_DOWN", "ROUND_HALF_UP"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union, Iterable
from datetime import datetime
from typing_extensions import Literal, Required, Annotated, TypedDict

from ...._utils import PropertyInfo

__all__ = [
    "CommitCreateParams",
    "AccessSchedule",
    "AccessScheduleScheduleItem",
    "InvoiceSchedule",
    "InvoiceScheduleRecurringSchedule",
    "InvoiceScheduleScheduleItem",
    "Specifier",
]


class CommitCreateParams(TypedDict, total=False):
    access_schedule: Required[AccessSchedule]
    """Schedule for distributing the commit to the customer.

    For "POSTPAID" commits only one schedule item is allowed and amount must match
    invoice_schedule total.
    """

    customer_id: Required[str]

    priority: Required[float]
    """
    If multiple credits or commits are applicable, the one with the lower priority
    will apply first.
    """

    product_id: Required[str]
    """ID of the fixed product associated with the commit.

    This is required because products are used to invoice the commit amount.
    """

    type: Required[Literal["PREPAID", "POSTPAID"]]

    applicable_contract_ids: List[str]
    """Which contract the commit applies to.

    If not provided, the commit applies to all contracts.
    """

    applicable_product_ids: List[str]
    """Which products the commit applies to.

    If applicable_product_ids, applicable_product_tags or specifiers are not
    provided, the commit applies to all products.
    """

    applicable_product_tags: List[str]
    """Which tags the commit applies to.

    If applicable_product_ids, applicable_product_tags or specifiers are not
    provided, the commit applies to all products.
    """

    custom_fields: Dict[str, str]

    description: str
    """Used only in UI/API. It is not exposed to end customers."""

    invoice_contract_id: str
    """The contract that this commit will be billed on.

    This is required for "POSTPAID" commits and for "PREPAID" commits unless there
    is no invoice schedule above (i.e., the commit is 'free').
    """

    invoice_schedule: InvoiceSchedule
    """
    Required for "POSTPAID" commits: the true up invoice will be generated at this
    time and only one schedule item is allowed; the total must match
    accesss_schedule amount. Optional for "PREPAID" commits: if not provided, this
    will be a "complimentary" commit with no invoice.
    """

    name: str
    """displayed on invoices"""

    netsuite_sales_order_id: str
    """This field's availability is dependent on your client's configuration."""

    rate_type: Literal["COMMIT_RATE", "LIST_RATE"]

    salesforce_opportunity_id: str
    """This field's availability is dependent on your client's configuration."""

    specifiers: Iterable[Specifier]
    """
    List of filters that determine what kind of customer usage draws down a commit
    or credit. A customer's usage needs to meet the condition of at least one of the
    specifiers to contribute to a commit's or credit's drawdown. This field cannot
    be used together with `applicable_product_ids` or `applicable_product_tags`.
    """

    uniqueness_key: str
    """Prevents the creation of duplicates.

    If a request to create a commit or credit is made with a uniqueness key that was
    previously used to create a commit or credit, a new record will not be created
    and the request will fail with a 409 error.
    """


class AccessScheduleScheduleItem(TypedDict, total=False):
    amount: Required[float]

    ending_before: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """RFC 3339 timestamp (exclusive)"""

    starting_at: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """RFC 3339 timestamp (inclusive)"""


class AccessSchedule(TypedDict, total=False):
    schedule_items: Required[Iterable[AccessScheduleScheduleItem]]

    credit_type_id: str
    """Defaults to USD (cents) if not passed"""


class InvoiceScheduleRecurringSchedule(TypedDict, total=False):
    amount_distribution: Required[Literal["DIVIDED", "DIVIDED_ROUNDED", "EACH"]]

    ending_before: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """RFC 3339 timestamp (exclusive)."""

    frequency: Required[Literal["MONTHLY", "QUARTERLY", "SEMI_ANNUAL", "ANNUAL"]]

    starting_at: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """RFC 3339 timestamp (inclusive)."""

    amount: float
    """Amount for the charge.

    Can be provided instead of unit_price and quantity. If amount is sent, the
    unit_price is assumed to be the amount and quantity is inferred to be 1.
    """

    quantity: float
    """Quantity for the charge.

    Will be multiplied by unit_price to determine the amount and must be specified
    with unit_price. If specified amount cannot be provided.
    """

    unit_price: float
    """Unit price for the charge.

    Will be multiplied by quantity to determine the amount and must be specified
    with quantity. If specified amount cannot be provided.
    """


class InvoiceScheduleScheduleItem(TypedDict, total=False):
    timestamp: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """timestamp of the scheduled event"""

    amount: float
    """Amount for the charge.

    Can be provided instead of unit_price and quantity. If amount is sent, the
    unit_price is assumed to be the amount and quantity is inferred to be 1.
    """

    quantity: float
    """Quantity for the charge.

    Will be multiplied by unit_price to determine the amount and must be specified
    with unit_price. If specified amount cannot be provided.
    """

    unit_price: float
    """Unit price for the charge.

    Will be multiplied by quantity to determine the amount and must be specified
    with quantity. If specified amount cannot be provided.
    """


class InvoiceSchedule(TypedDict, total=False):
    credit_type_id: str
    """Defaults to USD (cents) if not passed."""

    recurring_schedule: InvoiceScheduleRecurringSchedule
    """Enter the unit price and quantity for the charge or instead only send the
    amount.

    If amount is sent, the unit price is assumed to be the amount and quantity is
    inferred to be 1.
    """

    schedule_items: Iterable[InvoiceScheduleScheduleItem]
    """Either provide amount or provide both unit_price and quantity."""


class Specifier(TypedDict, total=False):
    presentation_group_values: Dict[str, str]

    pricing_group_values: Dict[str, str]

    product_id: str
    """
    If provided, the specifier will only apply to the product with the specified ID.
    """

    product_tags: List[str]
    """
    If provided, the specifier will only apply to products with all the specified
    tags.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .v1 import (
    V1Resource,
    AsyncV1Resource,
    V1ResourceWithRawResponse,
    AsyncV1ResourceWithRawResponse,
    V1ResourceWithStreamingResponse,
    AsyncV1ResourceWithStreamingResponse,
)
from .plans import (
    PlansResource,
    AsyncPlansResource,
    PlansResourceWithRawResponse,
    AsyncPlansResourceWithRawResponse,
    PlansResourceWithStreamingResponse,
    AsyncPlansResourceWithStreamingResponse,
)
from .usage import (
    UsageResource,
    AsyncUsageResource,
    UsageResourceWithRawResponse,
    AsyncUsageResourceWithRawResponse,
    UsageResourceWithStreamingResponse,
    AsyncUsageResourceWithStreamingResponse,
)
from .alerts import (
    AlertsResource,
    AsyncAlertsResource,
    AlertsResourceWithRawResponse,
    AsyncAlertsResourceWithRawResponse,
    AlertsResourceWithStreamingResponse,
    AsyncAlertsResourceWithStreamingResponse,
)
from .invoices import (
    InvoicesResource,
    AsyncInvoicesResource,
    InvoicesResourceWithRawResponse,
    AsyncInvoicesResourceWithRawResponse,
    InvoicesResourceWithStreamingResponse,
    AsyncInvoicesResourceWithStreamingResponse,
)
from .services import (
    ServicesResource,
    AsyncServicesResource,
    ServicesResourceWithRawResponse,
    AsyncServicesResourceWithRawResponse,
    ServicesResourceWithStreamingResponse,
    AsyncServicesResourceWithStreamingResponse,
)
from .contracts import (
    ContractsResource,
    AsyncContractsResource,
    ContractsResourceWithRawResponse,
    AsyncContractsResourceWithRawResponse,
    ContractsResourceWithStreamingResponse,
    AsyncContractsResourceWithStreamingResponse,
)
from .customers import (
    CustomersResource,
    AsyncCustomersResource,
    CustomersResourceWithRawResponse,
    AsyncCustomersResourceWithRawResponse,
    CustomersResourceWithStreamingResponse,
    AsyncCustomersResourceWithStreamingResponse,
)
from .audit_logs import (
    AuditLogsResource,
    AsyncAuditLogsResource,
    AuditLogsResourceWithRawResponse,
    AsyncAuditLogsResourceWithRawResponse,
    AuditLogsResourceWithStreamingResponse,
    AsyncAuditLogsResourceWithStreamingResponse,
)
from .dashboards import (
    DashboardsResource,
    AsyncDashboardsResource,
    DashboardsResourceWithRawResponse,
    AsyncDashboardsResourceWithRawResponse,
    DashboardsResourceWithStreamingResponse,
    AsyncDashboardsResourceWithStreamingResponse,
)
from .credit_grants import (
    CreditGrantsResource,
    AsyncCreditGrantsResource,
    CreditGrantsResourceWithRawResponse,
    AsyncCreditGrantsResourceWithRawResponse,
    CreditGrantsResourceWithStreamingResponse,
    AsyncCreditGrantsResourceWithStreamingResponse,
)
from .custom_fields import (
    CustomFieldsResource,
    AsyncCustomFieldsResource,
    CustomFieldsResourceWithRawResponse,
    AsyncCustomFieldsResourceWithRawResponse,
    CustomFieldsResourceWithStreamingResponse,
    AsyncCustomFieldsResourceWithStreamingResponse,
)
from .pricing_units import (
    PricingUnitsResource,
    AsyncPricingUnitsResource,
    PricingUnitsResourceWithRawResponse,
    AsyncPricingUnitsResourceWithRawResponse,
    PricingUnitsResourceWithStreamingResponse,
    AsyncPricingUnitsResourceWithStreamingResponse,
)
from .billable_metrics import (
    BillableMetricsResource,
    AsyncBillableMetricsResource,
    BillableMetricsResourceWithRawResponse,
    AsyncBillableMetricsResourceWithRawResponse,
    BillableMetricsResourceWithStreamingResponse,
    AsyncBillableMetricsResourceWithStreamingResponse,
)

__all__ = [
    "AlertsResource",
    "AsyncAlertsResource",
    "AlertsResourceWithRawResponse",
    "AsyncAlertsResourceWithRawResponse",
    "AlertsResourceWithStreamingResponse",
    "AsyncAlertsResourceWithStreamingResponse",
    "PlansResource",
    "AsyncPlansResource",
    "PlansResourceWithRawResponse",
    "AsyncPlansResourceWithRawResponse",
    "PlansResourceWithStreamingResponse",
    "AsyncPlansResourceWithStreamingResponse",
    "CreditGrantsResource",
    "AsyncCreditGrantsResource",
    "CreditGrantsResourceWithRawResponse",
    "AsyncCreditGrantsResourceWithRawResponse",
    "CreditGrantsResourceWithStreamingResponse",
    "AsyncCreditGrantsResourceWithStreamingResponse",
    "PricingUnitsResource",
    "AsyncPricingUnitsResource",
    "PricingUnitsResourceWithRawResponse",
    "AsyncPricingUnitsResourceWithRawResponse",
    "PricingUnitsResourceWithStreamingResponse",
    "AsyncPricingUnitsResourceWithStreamingResponse",
    "CustomersResource",
    "AsyncCustomersResource",
    "CustomersResourceWithRawResponse",
    "AsyncCustomersResourceWithRawResponse",
    "CustomersResourceWithStreamingResponse",
    "AsyncCustomersResourceWithStreamingResponse",
    "DashboardsResource",
    "AsyncDashboardsResource",
    "DashboardsResourceWithRawResponse",
    "AsyncDashboardsResourceWithRawResponse",
    "DashboardsResourceWithStreamingResponse",
    "AsyncDashboardsResourceWithStreamingResponse",
    "UsageResource",
    "AsyncUsageResource",
    "UsageResourceWithRawResponse",
    "AsyncUsageResourceWithRawResponse",
    "UsageResourceWithStreamingResponse",
    "AsyncUsageResourceWithStreamingResponse",
    "AuditLogsResource",
    "AsyncAuditLogsResource",
    "AuditLogsResourceWithRawResponse",
    "AsyncAuditLogsResourceWithRawResponse",
    "AuditLogsResourceWithStreamingResponse",
    "AsyncAuditLogsResourceWithStreamingResponse",
    "CustomFieldsResource",
    "AsyncCustomFieldsResource",
    "CustomFieldsResourceWithRawResponse",
    "AsyncCustomFieldsResourceWithRawResponse",
    "CustomFieldsResourceWithStreamingResponse",
    "AsyncCustomFieldsResourceWithStreamingResponse",
    "BillableMetricsResource",
    "AsyncBillableMetricsResource",
    "BillableMetricsResourceWithRawResponse",
    "AsyncBillableMetricsResourceWithRawResponse",
    "BillableMetricsResourceWithStreamingResponse",
    "AsyncBillableMetricsResourceWithStreamingResponse",
    "ServicesResource",
    "AsyncServicesResource",
    "ServicesResourceWithRawResponse",
    "AsyncServicesResourceWithRawResponse",
    "ServicesResourceWithStreamingResponse",
    "AsyncServicesResourceWithStreamingResponse",
    "InvoicesResource",
    "AsyncInvoicesResource",
    "InvoicesResourceWithRawResponse",
    "AsyncInvoicesResourceWithRawResponse",
    "InvoicesResourceWithStreamingResponse",
    "AsyncInvoicesResourceWithStreamingResponse",
    "ContractsResource",
    "AsyncContractsResource",
    "ContractsResourceWithRawResponse",
    "AsyncContractsResourceWithRawResponse",
    "ContractsResourceWithStreamingResponse",
    "AsyncContractsResourceWithStreamingResponse",
    "V1Resource",
    "AsyncV1Resource",
    "V1ResourceWithRawResponse",
    "AsyncV1ResourceWithRawResponse",
    "V1ResourceWithStreamingResponse",
    "AsyncV1ResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from metronome import Metronome, AsyncMetronome
from tests.utils import assert_matches_type
from metronome._utils import parse_datetime
from metronome.types.v1 import (
    UsageListResponse,
    UsageSearchResponse,
    UsageListWithGroupsResponse,
)
from metronome.pagination import SyncCursorPage, AsyncCursorPage

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestUsage:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_list(self, client: Metronome) -> None:
        usage = client.v1.usage.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
        )
        assert_matches_type(UsageListResponse, usage, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Metronome) -> None:
        usage = client.v1.usage.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
            next_page="next_page",
            billable_metrics=[
                {
                    "id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "group_by": {
                        "key": "key",
                        "values": ["x"],
                    },
                }
            ],
            customer_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        )
        assert_matches_type(UsageListResponse, usage, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Metronome) -> None:
        response = client.v1.usage.with_raw_response.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = response.parse()
        assert_matches_type(UsageListResponse, usage, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Metronome) -> None:
        with client.v1.usage.with_streaming_response.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = response.parse()
            assert_matches_type(UsageListResponse, usage, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_ingest(self, client: Metronome) -> None:
        usage = client.v1.usage.ingest()
        assert usage is None

    @parametrize
    def test_method_ingest_with_all_params(self, client: Metronome) -> None:
        usage = client.v1.usage.ingest(
            usage=[
                {
                    "customer_id": "team@example.com",
                    "event_type": "heartbeat",
                    "timestamp": "2021-01-01T00:00:00Z",
                    "transaction_id": "2021-01-01T00:00:00Z_cluster42",
                    "properties": {
                        "cluster_id": "bar",
                        "cpu_seconds": "bar",
                        "region": "bar",
                    },
                }
            ],
        )
        assert usage is None

    @parametrize
    def test_raw_response_ingest(self, client: Metronome) -> None:
        response = client.v1.usage.with_raw_response.ingest()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = response.parse()
        assert usage is None

    @parametrize
    def test_streaming_response_ingest(self, client: Metronome) -> None:
        with client.v1.usage.with_streaming_response.ingest() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = response.parse()
            assert usage is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_list_with_groups(self, client: Metronome) -> None:
        usage = client.v1.usage.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
        )
        assert_matches_type(SyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

    @parametrize
    def test_method_list_with_groups_with_all_params(self, client: Metronome) -> None:
        usage = client.v1.usage.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
            limit=1,
            next_page="next_page",
            current_period=True,
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            group_by={
                "key": "region",
                "values": ["US-East", "US-West", "EU-Central"],
            },
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
        )
        assert_matches_type(SyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

    @parametrize
    def test_raw_response_list_with_groups(self, client: Metronome) -> None:
        response = client.v1.usage.with_raw_response.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = response.parse()
        assert_matches_type(SyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

    @parametrize
    def test_streaming_response_list_with_groups(self, client: Metronome) -> None:
        with client.v1.usage.with_streaming_response.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = response.parse()
            assert_matches_type(SyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_search(self, client: Metronome) -> None:
        usage = client.v1.usage.search(
            transaction_ids=["2021-01-01T00:00:00Z_cluster42"],
        )
        assert_matches_type(UsageSearchResponse, usage, path=["response"])

    @parametrize
    def test_raw_response_search(self, client: Metronome) -> None:
        response = client.v1.usage.with_raw_response.search(
            transaction_ids=["2021-01-01T00:00:00Z_cluster42"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = response.parse()
        assert_matches_type(UsageSearchResponse, usage, path=["response"])

    @parametrize
    def test_streaming_response_search(self, client: Metronome) -> None:
        with client.v1.usage.with_streaming_response.search(
            transaction_ids=["2021-01-01T00:00:00Z_cluster42"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = response.parse()
            assert_matches_type(UsageSearchResponse, usage, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncUsage:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_list(self, async_client: AsyncMetronome) -> None:
        usage = await async_client.v1.usage.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
        )
        assert_matches_type(UsageListResponse, usage, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncMetronome) -> None:
        usage = await async_client.v1.usage.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
            next_page="next_page",
            billable_metrics=[
                {
                    "id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "group_by": {
                        "key": "key",
                        "values": ["x"],
                    },
                }
            ],
            customer_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        )
        assert_matches_type(UsageListResponse, usage, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.usage.with_raw_response.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = await response.parse()
        assert_matches_type(UsageListResponse, usage, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.usage.with_streaming_response.list(
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
            window_size="HOUR",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = await response.parse()
            assert_matches_type(UsageListResponse, usage, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_ingest(self, async_client: AsyncMetronome) -> None:
        usage = await async_client.v1.usage.ingest()
        assert usage is None

    @parametrize
    async def test_method_ingest_with_all_params(self, async_client: AsyncMetronome) -> None:
        usage = await async_client.v1.usage.ingest(
            usage=[
                {
                    "customer_id": "team@example.com",
                    "event_type": "heartbeat",
                    "timestamp": "2021-01-01T00:00:00Z",
                    "transaction_id": "2021-01-01T00:00:00Z_cluster42",
                    "properties": {
                        "cluster_id": "bar",
                        "cpu_seconds": "bar",
                        "region": "bar",
                    },
                }
            ],
        )
        assert usage is None

    @parametrize
    async def test_raw_response_ingest(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.usage.with_raw_response.ingest()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = await response.parse()
        assert usage is None

    @parametrize
    async def test_streaming_response_ingest(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.usage.with_streaming_response.ingest() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = await response.parse()
            assert usage is None

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_list_with_groups(self, async_client: AsyncMetronome) -> None:
        usage = await async_client.v1.usage.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
        )
        assert_matches_type(AsyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

    @parametrize
    async def test_method_list_with_groups_with_all_params(self, async_client: AsyncMetronome) -> None:
        usage = await async_client.v1.usage.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
            limit=1,
            next_page="next_page",
            current_period=True,
            ending_before=parse_datetime("2021-01-03T00:00:00Z"),
            group_by={
                "key": "region",
                "values": ["US-East", "US-West", "EU-Central"],
            },
            starting_on=parse_datetime("2021-01-01T00:00:00Z"),
        )
        assert_matches_type(AsyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

    @parametrize
    async def test_raw_response_list_with_groups(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.usage.with_raw_response.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = await response.parse()
        assert_matches_type(AsyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

    @parametrize
    async def test_streaming_response_list_with_groups(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.usage.with_streaming_response.list_with_groups(
            billable_metric_id="222796fd-d29c-429e-89b2-549fabda4ed6",
            customer_id="04ca7e72-4229-4a6e-ab11-9f7376fccbcb",
            window_size="HOUR",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = await response.parse()
            assert_matches_type(AsyncCursorPage[UsageListWithGroupsResponse], usage, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_search(self, async_client: AsyncMetronome) -> None:
        usage = await async_client.v1.usage.search(
            transaction_ids=["2021-01-01T00:00:00Z_cluster42"],
        )
        assert_matches_type(UsageSearchResponse, usage, path=["response"])

    @parametrize
    async def test_raw_response_search(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.usage.with_raw_response.search(
            transaction_ids=["2021-01-01T00:00:00Z_cluster42"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        usage = await response.parse()
        assert_matches_type(UsageSearchResponse, usage, path=["response"])

    @parametrize
    async def test_streaming_response_search(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.usage.with_streaming_response.search(
            transaction_ids=["2021-01-01T00:00:00Z_cluster42"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            usage = await response.parse()
            assert_matches_type(UsageSearchResponse, usage, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Union, Optional
from datetime import datetime
from typing_extensions import Literal, TypeAlias

from ..._models import BaseModel
from .schedule_duration import ScheduleDuration

__all__ = [
    "Credit",
    "Product",
    "Contract",
    "HierarchyConfiguration",
    "HierarchyConfigurationChildAccess",
    "HierarchyConfigurationChildAccessCommitHierarchyChildAccessAll",
    "HierarchyConfigurationChildAccessCommitHierarchyChildAccessNone",
    "HierarchyConfigurationChildAccessCommitHierarchyChildAccessContractIDs",
    "Ledger",
    "LedgerCreditSegmentStartLedgerEntry",
    "LedgerCreditAutomatedInvoiceDeductionLedgerEntry",
    "LedgerCreditExpirationLedgerEntry",
    "LedgerCreditCanceledLedgerEntry",
    "LedgerCreditCreditedLedgerEntry",
    "LedgerCreditManualLedgerEntry",
    "LedgerCreditSeatBasedAdjustmentLedgerEntry",
    "Specifier",
]


class Product(BaseModel):
    id: str

    name: str


class Contract(BaseModel):
    id: str


class HierarchyConfigurationChildAccessCommitHierarchyChildAccessAll(BaseModel):
    type: Literal["ALL"]


class HierarchyConfigurationChildAccessCommitHierarchyChildAccessNone(BaseModel):
    type: Literal["NONE"]


class HierarchyConfigurationChildAccessCommitHierarchyChildAccessContractIDs(BaseModel):
    contract_ids: List[str]

    type: Literal["CONTRACT_IDS"]


HierarchyConfigurationChildAccess: TypeAlias = Union[
    HierarchyConfigurationChildAccessCommitHierarchyChildAccessAll,
    HierarchyConfigurationChildAccessCommitHierarchyChildAccessNone,
    HierarchyConfigurationChildAccessCommitHierarchyChildAccessContractIDs,
]


class HierarchyConfiguration(BaseModel):
    child_access: HierarchyConfigurationChildAccess


class LedgerCreditSegmentStartLedgerEntry(BaseModel):
    amount: float

    segment_id: str

    timestamp: datetime

    type: Literal["CREDIT_SEGMENT_START"]


class LedgerCreditAutomatedInvoiceDeductionLedgerEntry(BaseModel):
    amount: float

    invoice_id: str

    segment_id: str

    timestamp: datetime

    type: Literal["CREDIT_AUTOMATED_INVOICE_DEDUCTION"]

    contract_id: Optional[str] = None


class LedgerCreditExpirationLedgerEntry(BaseModel):
    amount: float

    segment_id: str

    timestamp: datetime

    type: Literal["CREDIT_EXPIRATION"]


class LedgerCreditCanceledLedgerEntry(BaseModel):
    amount: float

    invoice_id: str

    segment_id: str

    timestamp: datetime

    type: Literal["CREDIT_CANCELED"]

    contract_id: Optional[str] = None


class LedgerCreditCreditedLedgerEntry(BaseModel):
    amount: float

    invoice_id: str

    segment_id: str

    timestamp: datetime

    type: Literal["CREDIT_CREDITED"]

    contract_id: Optional[str] = None


class LedgerCreditManualLedgerEntry(BaseModel):
    amount: float

    reason: str

    timestamp: datetime

    type: Literal["CREDIT_MANUAL"]


class LedgerCreditSeatBasedAdjustmentLedgerEntry(BaseModel):
    amount: float

    segment_id: str

    timestamp: datetime

    type: Literal["CREDIT_SEAT_BASED_ADJUSTMENT"]


Ledger: TypeAlias = Union[
    LedgerCreditSegmentStartLedgerEntry,
    LedgerCreditAutomatedInvoiceDeductionLedgerEntry,
    LedgerCreditExpirationLedgerEntry,
    LedgerCreditCanceledLedgerEntry,
    LedgerCreditCreditedLedgerEntry,
    LedgerCreditManualLedgerEntry,
    LedgerCreditSeatBasedAdjustmentLedgerEntry,
]


class Specifier(BaseModel):
    presentation_group_values: Optional[Dict[str, str]] = None

    pricing_group_values: Optional[Dict[str, str]] = None

    product_id: Optional[str] = None
    """
    If provided, the specifier will only apply to the product with the specified ID.
    """

    product_tags: Optional[List[str]] = None
    """
    If provided, the specifier will only apply to products with all the specified
    tags.
    """


class Credit(BaseModel):
    id: str

    product: Product

    type: Literal["CREDIT"]

    access_schedule: Optional[ScheduleDuration] = None
    """The schedule that the customer will gain access to the credits."""

    applicable_contract_ids: Optional[List[str]] = None

    applicable_product_ids: Optional[List[str]] = None

    applicable_product_tags: Optional[List[str]] = None

    balance: Optional[float] = None
    """The current balance of the credit or commit.

    This balance reflects the amount of credit or commit that the customer has
    access to use at this moment - thus, expired and upcoming credit or commit
    segments contribute 0 to the balance. The balance will match the sum of all
    ledger entries with the exception of the case where the sum of negative manual
    ledger entries exceeds the positive amount remaining on the credit or commit -
    in that case, the balance will be 0. All manual ledger entries associated with
    active credit or commit segments are included in the balance, including
    future-dated manual ledger entries.
    """

    contract: Optional[Contract] = None

    custom_fields: Optional[Dict[str, str]] = None

    description: Optional[str] = None

    hierarchy_configuration: Optional[HierarchyConfiguration] = None
    """Optional configuration for credit hierarchy access control"""

    ledger: Optional[List[Ledger]] = None
    """A list of ordered events that impact the balance of a credit.

    For example, an invoice deduction or an expiration.
    """

    name: Optional[str] = None

    netsuite_sales_order_id: Optional[str] = None
    """This field's availability is dependent on your client's configuration."""

    priority: Optional[float] = None
    """
    If multiple credits or commits are applicable, the one with the lower priority
    will apply first.
    """

    rate_type: Optional[Literal["COMMIT_RATE", "LIST_RATE"]] = None

    salesforce_opportunity_id: Optional[str] = None
    """This field's availability is dependent on your client's configuration."""

    specifiers: Optional[List[Specifier]] = None
    """
    List of filters that determine what kind of customer usage draws down a commit
    or credit. A customer's usage needs to meet the condition of at least one of the
    specifiers to contribute to a commit's or credit's drawdown.
    """

    uniqueness_key: Optional[str] = None
    """Prevents the creation of duplicates.

    If a request to create a commit or credit is made with a uniqueness key that was
    previously used to create a commit or credit, a new record will not be created
    and the request will fail with a 409 error.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .tier import Tier as Tier
from .property_filter import PropertyFilter as PropertyFilter
from .base_usage_filter import BaseUsageFilter as BaseUsageFilter
from .event_type_filter import EventTypeFilter as EventTypeFilter



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional

from ..._models import BaseModel
from ..shared.credit_type_data import CreditTypeData

__all__ = ["PlanDetail", "CreditGrant", "Minimum", "OverageRate"]


class CreditGrant(BaseModel):
    amount_granted: float

    amount_granted_credit_type: CreditTypeData

    amount_paid: float

    amount_paid_credit_type: CreditTypeData

    effective_duration: float

    name: str

    priority: str

    send_invoice: bool

    reason: Optional[str] = None

    recurrence_duration: Optional[float] = None

    recurrence_interval: Optional[float] = None


class Minimum(BaseModel):
    credit_type: CreditTypeData

    name: str

    start_period: float
    """Used in price ramps.

    Indicates how many billing periods pass before the charge applies.
    """

    value: float


class OverageRate(BaseModel):
    credit_type: CreditTypeData

    fiat_credit_type: CreditTypeData

    start_period: float
    """Used in price ramps.

    Indicates how many billing periods pass before the charge applies.
    """

    to_fiat_conversion_factor: float


class PlanDetail(BaseModel):
    id: str

    custom_fields: Dict[str, str]

    name: str

    credit_grants: Optional[List[CreditGrant]] = None

    description: Optional[str] = None

    minimums: Optional[List[Minimum]] = None

    overage_rates: Optional[List[OverageRate]] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from metronome import Metronome, AsyncMetronome
from tests.utils import assert_matches_type
from metronome._utils import parse_datetime
from metronome.types.v1.customers import (
    CreditListResponse,
    CreditCreateResponse,
    CreditUpdateEndDateResponse,
)

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestCredits:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_create(self, client: Metronome) -> None:
        credit = client.v1.customers.credits.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ]
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
        )
        assert_matches_type(CreditCreateResponse, credit, path=["response"])

    @parametrize
    def test_method_create_with_all_params(self, client: Metronome) -> None:
        credit = client.v1.customers.credits.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ],
                "credit_type_id": "2714e483-4ff1-48e4-9e25-ac732e8f24f2",
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
            applicable_contract_ids=["string"],
            applicable_product_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
            applicable_product_tags=["string"],
            custom_fields={"foo": "string"},
            description="description",
            name="My Credit",
            netsuite_sales_order_id="netsuite_sales_order_id",
            rate_type="COMMIT_RATE",
            salesforce_opportunity_id="salesforce_opportunity_id",
            specifiers=[
                {
                    "presentation_group_values": {"foo": "string"},
                    "pricing_group_values": {"foo": "string"},
                    "product_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "product_tags": ["string"],
                }
            ],
            uniqueness_key="x",
        )
        assert_matches_type(CreditCreateResponse, credit, path=["response"])

    @parametrize
    def test_raw_response_create(self, client: Metronome) -> None:
        response = client.v1.customers.credits.with_raw_response.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ]
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        credit = response.parse()
        assert_matches_type(CreditCreateResponse, credit, path=["response"])

    @parametrize
    def test_streaming_response_create(self, client: Metronome) -> None:
        with client.v1.customers.credits.with_streaming_response.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ]
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            credit = response.parse()
            assert_matches_type(CreditCreateResponse, credit, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_list(self, client: Metronome) -> None:
        credit = client.v1.customers.credits.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )
        assert_matches_type(CreditListResponse, credit, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Metronome) -> None:
        credit = client.v1.customers.credits.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            covering_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            effective_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            include_archived=True,
            include_balance=True,
            include_contract_credits=True,
            include_ledgers=True,
            limit=1,
            next_page="next_page",
            starting_at=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(CreditListResponse, credit, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Metronome) -> None:
        response = client.v1.customers.credits.with_raw_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        credit = response.parse()
        assert_matches_type(CreditListResponse, credit, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Metronome) -> None:
        with client.v1.customers.credits.with_streaming_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            credit = response.parse()
            assert_matches_type(CreditListResponse, credit, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_update_end_date(self, client: Metronome) -> None:
        credit = client.v1.customers.credits.update_end_date(
            access_ending_before=parse_datetime("2020-01-01T00:00:00.000Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )
        assert_matches_type(CreditUpdateEndDateResponse, credit, path=["response"])

    @parametrize
    def test_raw_response_update_end_date(self, client: Metronome) -> None:
        response = client.v1.customers.credits.with_raw_response.update_end_date(
            access_ending_before=parse_datetime("2020-01-01T00:00:00.000Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        credit = response.parse()
        assert_matches_type(CreditUpdateEndDateResponse, credit, path=["response"])

    @parametrize
    def test_streaming_response_update_end_date(self, client: Metronome) -> None:
        with client.v1.customers.credits.with_streaming_response.update_end_date(
            access_ending_before=parse_datetime("2020-01-01T00:00:00.000Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            credit = response.parse()
            assert_matches_type(CreditUpdateEndDateResponse, credit, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncCredits:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_create(self, async_client: AsyncMetronome) -> None:
        credit = await async_client.v1.customers.credits.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ]
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
        )
        assert_matches_type(CreditCreateResponse, credit, path=["response"])

    @parametrize
    async def test_method_create_with_all_params(self, async_client: AsyncMetronome) -> None:
        credit = await async_client.v1.customers.credits.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ],
                "credit_type_id": "2714e483-4ff1-48e4-9e25-ac732e8f24f2",
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
            applicable_contract_ids=["string"],
            applicable_product_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
            applicable_product_tags=["string"],
            custom_fields={"foo": "string"},
            description="description",
            name="My Credit",
            netsuite_sales_order_id="netsuite_sales_order_id",
            rate_type="COMMIT_RATE",
            salesforce_opportunity_id="salesforce_opportunity_id",
            specifiers=[
                {
                    "presentation_group_values": {"foo": "string"},
                    "pricing_group_values": {"foo": "string"},
                    "product_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "product_tags": ["string"],
                }
            ],
            uniqueness_key="x",
        )
        assert_matches_type(CreditCreateResponse, credit, path=["response"])

    @parametrize
    async def test_raw_response_create(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.credits.with_raw_response.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ]
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        credit = await response.parse()
        assert_matches_type(CreditCreateResponse, credit, path=["response"])

    @parametrize
    async def test_streaming_response_create(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.credits.with_streaming_response.create(
            access_schedule={
                "schedule_items": [
                    {
                        "amount": 1000,
                        "ending_before": parse_datetime("2020-02-01T00:00:00.000Z"),
                        "starting_at": parse_datetime("2020-01-01T00:00:00.000Z"),
                    }
                ]
            },
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            priority=100,
            product_id="f14d6729-6a44-4b13-9908-9387f1918790",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            credit = await response.parse()
            assert_matches_type(CreditCreateResponse, credit, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_list(self, async_client: AsyncMetronome) -> None:
        credit = await async_client.v1.customers.credits.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )
        assert_matches_type(CreditListResponse, credit, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncMetronome) -> None:
        credit = await async_client.v1.customers.credits.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
            covering_date=parse_datetime("2019-12-27T18:11:19.117Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            effective_before=parse_datetime("2019-12-27T18:11:19.117Z"),
            include_archived=True,
            include_balance=True,
            include_contract_credits=True,
            include_ledgers=True,
            limit=1,
            next_page="next_page",
            starting_at=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(CreditListResponse, credit, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.credits.with_raw_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        credit = await response.parse()
        assert_matches_type(CreditListResponse, credit, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.credits.with_streaming_response.list(
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            credit = await response.parse()
            assert_matches_type(CreditListResponse, credit, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_update_end_date(self, async_client: AsyncMetronome) -> None:
        credit = await async_client.v1.customers.credits.update_end_date(
            access_ending_before=parse_datetime("2020-01-01T00:00:00.000Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )
        assert_matches_type(CreditUpdateEndDateResponse, credit, path=["response"])

    @parametrize
    async def test_raw_response_update_end_date(self, async_client: AsyncMetronome) -> None:
        response = await async_client.v1.customers.credits.with_raw_response.update_end_date(
            access_ending_before=parse_datetime("2020-01-01T00:00:00.000Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        credit = await response.parse()
        assert_matches_type(CreditUpdateEndDateResponse, credit, path=["response"])

    @parametrize
    async def test_streaming_response_update_end_date(self, async_client: AsyncMetronome) -> None:
        async with async_client.v1.customers.credits.with_streaming_response.update_end_date(
            access_ending_before=parse_datetime("2020-01-01T00:00:00.000Z"),
            credit_id="6162d87b-e5db-4a33-b7f2-76ce6ead4e85",
            customer_id="13117714-3f05-48e5-a6e9-a66093f13b4d",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            credit = await response.parse()
            assert_matches_type(CreditUpdateEndDateResponse, credit, path=["response"])

        assert cast(Any, response.is_closed) is True
